// Copyright (C) 2010  Internet Systems Consortium, Inc. ("ISC")
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
// REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
// INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
// OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
// PERFORMANCE OF THIS SOFTWARE.

#include <dns/name.h>
#include <dns/rrclass.h>

#include <datasrc/zonetable.h>
#include <datasrc/rbtree.h>

using namespace std;
using namespace isc::dns;

namespace isc {
namespace datasrc {

struct MemoryZone::MemoryZoneImpl {
    MemoryZoneImpl(const RRClass& zone_class, const Name& origin) :
        zone_class_(zone_class), origin_(origin)
    {}
    RRClass zone_class_;
    Name origin_;
};

MemoryZone::MemoryZone(const RRClass& zone_class, const Name& origin) :
    impl_(new MemoryZoneImpl(zone_class, origin))
{
}

MemoryZone::~MemoryZone() {
    delete impl_;
}

const Name&
MemoryZone::getOrigin() const {
    return (impl_->origin_);
}

const RRClass&
MemoryZone::getClass() const {
    return (impl_->zone_class_);
}

Zone::FindResult
MemoryZone::find(const Name&, const RRType&) const {
    // This is a tentative implementation that always returns NXDOMAIN.
    return (FindResult(NXDOMAIN, RRsetPtr()));
}

struct ZoneTable::ZoneTableImpl {
    typedef RBTree<Zone> ZoneTree;
    typedef RBNode<Zone> ZoneNode;
    ZoneTree zones;

    result::Result addZone(ZonePtr zone) {
        // Sanity check
        if (!zone) {
            isc_throw(InvalidParameter,
                      "Null pointer is passed to ZoneTable::addZone()");
        }

        // Get the node where we put the zone
        ZoneNode* node(NULL);
        switch (zones.insert(zone->getOrigin(), &node)) {
            // This is OK
            case ZoneTree::SUCCEED:
            case ZoneTree::ALREADYEXIST:
                break;
            // Can Not Happen
            /*
             * This should generate no code if the implementation is correct
             * (since the compiler has complete code of RBTree right now, it
             * should see it can return only these two values). It is here
             * to catch programmer errors.
             */
            default:
                isc_throw(AssertError,
                    "RBTree<Zone>::insert returned unexpected result");
        }
        // Can Not Happen
        if (!node) {
            isc_throw(AssertError,
                "RBTree<Zone>::insert gave NULL pointer");
        }

        // Is it empty? We either just created it or it might be nonterminal
        if (node->isEmpty()) {
            node->setData(zone);
            return (result::SUCCESS);
        } else { // There's something there already
            return (result::EXIST);
        }
    }
};

ZoneTable::ZoneTable() : impl_(new ZoneTableImpl)
{}

ZoneTable::~ZoneTable() {
    delete impl_;
}

result::Result
ZoneTable::addZone(ZonePtr zone) {
    return (impl_->addZone(zone));
}

result::Result
ZoneTable::removeZone(const Name&) {
    // TODO Implement
    isc_throw(AssertError, "Not implemented");
}

ZoneTable::FindResult
ZoneTable::findZone(const Name& name) const {
    // Inefficient internal loop to find a longest match.
    // This will be replaced with a single call to more intelligent backend.
    for (int i = 0; i < name.getLabelCount(); ++i) {
        Name matchname(name.split(i));
        ZoneTableImpl::ZoneMap::const_iterator found =
            impl_->zones.find(matchname);
        if (found != impl_->zones.end()) {
            return (FindResult(i == 0 ? result::SUCCESS :
                               result::PARTIALMATCH, (*found).second));
        }
    }
    return (FindResult(result::NOTFOUND, ConstZonePtr()));
}
} // end of namespace datasrc
} // end of namespace isc
